/*-------------------------------------------------------------------------
 *
 * nwprepare.c
 *    - generates the directory structure on the destination server
 *    - generates the version file for PostgreSQL
 *	  - generates fmgroids.h and fmgrtab.c for NetWare
 *	    fmgroids.h and fmgrtab.c is based on include/catalog/pg_proc.h
 *    - generates postgres.bki for NetWare
 *    - checks for duplicate "analyze.c" and renames one to canalyze.c
 *
 * Portions Copyright (c) 1996-2002, PostgreSQL Global Development Group
 *
 * $Id: nwprepare.c v 0.9.1 2002/03/18 22:22:00 Ulrich Neumann Exp $
 *
 * NOTES
 *	  this file is only for the NetWare version of PostgreSQL. No other
 *	  Version will need this file! We need it because NetWare doesn't
 *    have a shell like bash! and SED (stream editor)
 *
 *-------------------------------------------------------------------------
 */

#include <stdio.h>
#include <stdlib.h>
#include <stat.h>
#include <fcntl.h>
#include <string.h>

#define _IS_NWPREPARE_
#include "pg_config.h"

/* define the environment variables for the template files here */
int pguid				= 1;
int encoding			= 5; /* UNICODE */
int	namedatalen			= 32;
int index_max_keys		= INDEX_MAX_KEYS;
int def_attstattarget	= DEFAULT_ATTSTATTARGET;

/* function prototypes */
int		main (int argc, char *argv[]);
void	RowSplitter (char* cvstring);
void	WriteToken (char* TokenString);
void	nwgenbki (FILE *bkihandle, char* inputfile);

/* global variables */
int		fmgrcounter;
int		defvalue;
char	defname[129];
int		defnumber;
int		defbool1;
int		defbool2;
int		oidcounter;

FILE	*fmgrhandle1;
FILE	*fmgrhandle2;
FILE	*fmgrhandletemp;

/* Temporary data holder */
typedef struct pgrow *PGROWPTR;

struct pgrow
	{
	int		number;
	char	name[40];
	int		value;
	int		db1;
	int		db2;
	struct pgrow *next;
	};

PGROWPTR	pghashtable	= NULL;
PGROWPTR	newpghash	= NULL;

int
main (int argc, char *argv[])
	{
	int			fd			= 0;
	int			stringloop	= 0;
	int			tempnumber	= 0;
	char		filename[256];
	char		bkiinputfile[256];
	char		testlinestring[512];
	char		testchar[1];
	char		defboolstr1[6];
	char		defboolstr2[6];
	char		*s;
	PGROWPTR	pghashtemp	= NULL;
	PGROWPTR	pghashtemp2	= NULL;
	FILE		*bkifile;

	/* generate directory structure				*/
	mkdir (NETWARE_DEST_PATH, 0);
	mkdir (NETWARE_DEST_PATH "/data", 0);
	mkdir (NETWARE_DEST_PATH "/data/base", 0);
	mkdir (NETWARE_DEST_PATH "/data/base/1", 0);
	mkdir (NETWARE_DEST_PATH "/data/global", 0);
	mkdir (NETWARE_DEST_PATH "/data/pg_xlog", 0);
	mkdir (NETWARE_DEST_PATH "/data/pg_clog", 0);
	mkdir (NETWARE_DEST_PATH "/share", 0);

	/* generate version files					*/
	fd = open(NETWARE_DEST_PATH "/data/PG_VERSION", O_CREAT | O_TRUNC | O_WRONLY, 0);
	write (fd, PG_VERSION, sizeof (PG_VERSION));
	close (fd);

	fd = open(NETWARE_DEST_PATH "/data/base/1/PG_VERSION", O_CREAT | O_TRUNC | O_WRONLY, 0);
	write (fd, PG_VERSION, sizeof (PG_VERSION));
	close (fd);

	/* check for analyze.c and rename it to canalyze.c if exists */
	sprintf (filename, "../../commands/analyze.c");
	if (access(filename, F_OK) == 0)
		rename (filename, (char *) "../../commands/canalyze.c");

	/* Copy conf files to the correct place */
	printf ("Copy pg_hba.conf...\n");
	fmgrhandletemp	= fopen (NETWARE_DEST_PATH "/data/pg_hba.conf", "w");
	fd = open ("../../libpq/pg_hba.conf.sample", O_RDONLY, 0);
	while (read (fd, testchar, 1) > 0)
		{
		if (testchar[0] == 13)
			{
			memmove (testlinestring + stringloop, "\0", 1);
			read (fd, testchar , 1);
			fprintf (fmgrhandletemp, "%s\n", testlinestring);
			stringloop = 0;
			}
			else
			{
			memmove (testlinestring + stringloop, testchar, 1);
			stringloop++;
			}
		}
	close (fd);
	fclose (fmgrhandletemp);

	printf ("Copy pg_ident.conf...\n");
	fmgrhandletemp	= fopen (NETWARE_DEST_PATH "/data/pg_ident.conf", "w");
	fd = open ("../../libpq/pg_ident.conf.sample", O_RDONLY, 0);
	while (read (fd, testchar, 1) > 0)
		{
		if (testchar[0] == 13)
			{
			memmove (testlinestring + stringloop, "\0", 1);
			read (fd, testchar , 1);
			fprintf (fmgrhandletemp, "%s\n", testlinestring);
			stringloop = 0;
			}
			else
			{
			memmove (testlinestring + stringloop, testchar, 1);
			stringloop++;
			}
		}
	close (fd);
	fclose (fmgrhandletemp);

	printf ("Copy postgresql.conf...\n");
	fmgrhandletemp	= fopen (NETWARE_DEST_PATH "/data/postgresql.conf", "w");
	fd = open ("../../utils/misc/postgresql.conf.sample", O_RDONLY, 0);
	while (read (fd, testchar, 1) > 0)
		{
		if (testchar[0] == 13)
			{
			memmove (testlinestring + stringloop, "\0", 1);
			read (fd, testchar , 1);
			fprintf (fmgrhandletemp, "%s\n", testlinestring);
			stringloop = 0;
			}
			else
			{
			memmove (testlinestring + stringloop, testchar, 1);
			stringloop++;
			}
		}
	close (fd);
	fclose (fmgrhandletemp);

	/* initialie hashtable */
	if (pghashtable = malloc (sizeof(struct pgrow)))
		{
		pghashtable->number	= NULL;
		pghashtable->next	= NULL;
		}

	/* generate fmgroids.h and fmgrtab.c		*/
	fmgrhandle1		= fopen("../../utils/fmgroids.h", "w");
	fmgrhandle2		= fopen("../../utils/fmgrtab.c", "w");
	fmgrhandletemp	= fopen (NETWARE_DEST_PATH "/fmgrtemp.c", "w");

	/* generate head of fmgroids.h				*/
	printf ("Genrate fmgroids.h head...\n");
	fprintf (fmgrhandle1, "/*-------------------------------------------------------------------------\n");
	fprintf (fmgrhandle1, " *\n");
	fprintf (fmgrhandle1, " * fmgroids.h\n");
	fprintf (fmgrhandle1, " *    Macros that define the OIDs of built-in functions.\n");
	fprintf (fmgrhandle1, " *\n");
	fprintf (fmgrhandle1, " * These macros can be used to avoid a catalog lookup when a specific\n");
	fprintf (fmgrhandle1, " * fmgr-callable function needs to be referenced.\n");
	fprintf (fmgrhandle1, " *\n");
	fprintf (fmgrhandle1, " * Portions Copyright (c) 1996-2001, PostgreSQL Global Development Group\n");
	fprintf (fmgrhandle1, " * Portions Copyright (c) 1994, Regents of the University of California\n");
	fprintf (fmgrhandle1, " *\n");
	fprintf (fmgrhandle1, " * NOTES\n");
	fprintf (fmgrhandle1, " *	******************************\n");
	fprintf (fmgrhandle1, " *	*** DO NOT EDIT THIS FILE! ***\n");
	fprintf (fmgrhandle1, " *	******************************\n");
	fprintf (fmgrhandle1, " *\n");
	fprintf (fmgrhandle1, " *	It has been GENERATED by Gen_fmgrtab.sh\n");
	fprintf (fmgrhandle1, " *	from ../../../src/include/catalog/pg_proc.h\n");
	fprintf (fmgrhandle1, " *\n");
	fprintf (fmgrhandle1, " *-------------------------------------------------------------------------\n");
	fprintf (fmgrhandle1, " */\n");
	fprintf (fmgrhandle1, "#ifndef	FMGROIDS_H\n");
	fprintf (fmgrhandle1, "#define FMGROIDS_H\n\n");
	fprintf (fmgrhandle1, "/*\n");
	fprintf (fmgrhandle1, " *	Constant macros for the OIDs of entries in pg_proc.\n");
	fprintf (fmgrhandle1, " *\n");
	fprintf (fmgrhandle1, " *	NOTE: macros are named after the prosrc value, ie the actual C name\n");
	fprintf (fmgrhandle1, " *	of the implementing function, not the proname which may be overloaded.\n");
	fprintf (fmgrhandle1, " *	For example, we want to be able to assign different macro names to both\n");
	fprintf (fmgrhandle1, " *	char_text() and int4_text() even though these both appear with proname\n");
	fprintf (fmgrhandle1, " *	'text'.  If the same C function appears in more than one pg_proc entry,\n");
	fprintf (fmgrhandle1, " *	its equivalent macro will be defined with the lowest OID among those\n");
	fprintf (fmgrhandle1, " *	entries.\n");
	fprintf (fmgrhandle1, " */\n");

	/* generate head of fmgrtab.c				*/
	printf ("Genrate fmgrtab.c head...\n");
	fprintf (fmgrhandle2, "/*-------------------------------------------------------------------------\n");
	fprintf (fmgrhandle2, " *\n");
	fprintf (fmgrhandle2, " * fmgrtab.c\n");
	fprintf (fmgrhandle2, " *    The function manager's table of internal functions.\n");
	fprintf (fmgrhandle2, " *\n");
	fprintf (fmgrhandle2, " * Portions Copyright (c) 1996-2001, PostgreSQL Global Development Group\n");
	fprintf (fmgrhandle2, " * Portions Copyright (c) 1994, Regents of the University of California\n");
	fprintf (fmgrhandle2, " *\n");
	fprintf (fmgrhandle2, " * NOTES\n");
	fprintf (fmgrhandle2, " *\n");
	fprintf (fmgrhandle2, " *	******************************\n");
	fprintf (fmgrhandle2, " *	*** DO NOT EDIT THIS FILE! ***\n");
	fprintf (fmgrhandle2, " *	******************************\n");
	fprintf (fmgrhandle2, " *\n");
	fprintf (fmgrhandle2, " *	It has been GENERATED by Gen_fmgrtab.sh\n");
	fprintf (fmgrhandle2, " *	from ../../../src/include/catalog/pg_proc.h\n");
	fprintf (fmgrhandle2, " *\n");
	fprintf (fmgrhandle2, " *-------------------------------------------------------------------------\n");
	fprintf (fmgrhandle2, " */\n\n");
	fprintf (fmgrhandle2, "#include \"postgres.h\"\n\n");
	fprintf (fmgrhandle2, "#include \"utils/fmgrtab.h\"\n\n");
	
	fd = open ("../../../include/catalog/pg_proc.h", O_RDONLY, 0);
	fmgrcounter	= 0;
	printf ("Read pg_proc.h...\n");

	while (read (fd, testchar, 1) > 0)		/* read line of inputfile */
		{
		if (testchar[0] == 13)
			{
			memmove (testlinestring + stringloop, "\0", 1);
			read (fd, testchar , 1);
			RowSplitter (testlinestring);
			stringloop = 0;
			}
			else
			{
			memmove (testlinestring + stringloop, testchar, 1);
			stringloop++;
			}
		}
	close (fd);

	/* Generate files from hashtable */
	printf ("Genrate sorted content of fmgroids.h...\n");
	pghashtemp		= pghashtable;
	while (pghashtemp)
		{
		if (pghashtemp->number > 0)
			{
			fprintf (fmgrhandle2, "extern Datum %s (PG_FUNCTION_ARGS);\n", &pghashtemp->name[0]);

			if (pghashtemp->db1) sprintf (defboolstr1, "true");
				else sprintf (defboolstr1, "false");
			if (pghashtemp->db2) sprintf (defboolstr2, "true");
				else sprintf (defboolstr2, "false");

			fprintf (fmgrhandletemp, "  { %d, \"%s\", %d, %s, %s, %s },\n", pghashtemp->number, &pghashtemp->name[0], pghashtemp->value, defboolstr1, defboolstr2, &pghashtemp->name[0]);

			s = &pghashtemp->name[0];
			for (; *s; s++)	if ((*s >= 97) && (*s <= 122)) *s -= 32;	/* convert to uppercase */
			pghashtemp2	= pghashtable;
			tempnumber	= 0;
			while (pghashtemp2)
				{
				if (pghashtemp2->number > 0)
					{
					if ((strncmp (&pghashtemp->name[0], &pghashtemp2->name[0], strlen (&pghashtemp->name[0])) == 0) &&
						(strncmp (&pghashtemp->name[0], &pghashtemp2->name[0], strlen (&pghashtemp2->name[0])) == 0))
						{
						tempnumber = pghashtemp2->number;
						break;
						}
					}
				pghashtemp2 = pghashtemp2->next;
				}

			if (pghashtemp->number == tempnumber)
				fprintf (fmgrhandle1, "#define F_%s %d\n", &pghashtemp->name[0], pghashtemp->number);
			}
		pghashtemp = pghashtemp->next;
		}

	printf ("Genrate fmgroids.h foot...\n");
	fprintf (fmgrhandle1, "\n");
	fprintf (fmgrhandle1, "#endif	/* FMGROIDS_H */\n");

	printf ("Genrate content of fmgrtemp.c\n");
	fprintf (fmgrhandle2, "\nconst FmgrBuiltin fmgr_builtins[] = {\n");
	fclose (fmgrhandletemp);

	fd = open (NETWARE_DEST_PATH "/fmgrtemp.c", O_RDONLY, 0);

	while (read (fd, testchar, 1) > 0)		/* read line of inputfile */
		{
		if (testchar[0] == 13)
			{
			memmove (testlinestring + stringloop, "\0", 1);
			read (fd, testchar , 1);
			fprintf (fmgrhandle2, "%s\n", testlinestring);
			stringloop = 0;
			}
			else
			{
			memmove (testlinestring + stringloop, testchar, 1);
			stringloop++;
			}
		}
	close (fd);
	remove (NETWARE_DEST_PATH "/fmgrtemp.c");

	printf ("Genrate fmgrtemp.c foot...\n");
	fprintf (fmgrhandle2, "  /* dummy entry is easier than getting rid of comma after last real one */\n");
	fprintf (fmgrhandle2, "  /* (not that there has ever been anything wrong with *having* a\n");
	fprintf (fmgrhandle2, "     comma after the last field in an array initializer) */\n");
	fprintf (fmgrhandle2, "  { 0, NULL, 0, false, false, (PGFunction) NULL }\n");
	fprintf (fmgrhandle2, "};\n\n");
	fprintf (fmgrhandle2, "/* Note fmgr_nbuiltins excludes the dummy entry */\n");
	fprintf (fmgrhandle2, "const int fmgr_nbuiltins = (sizeof(fmgr_builtins) / sizeof(FmgrBuiltin)) - 1;\n");

	fclose (fmgrhandle2);
	fclose (fmgrhandle1);

	/* generate postgres.description */
//	printf ("Genrate postgres.description...\n");
//	bkifile = fopen (NETWARE_DEST_PATH "/PREFIX/Share/postgres.description", "w");
//	fclose (bkifile);

	/* generate postgres.bki */
	bkifile = fopen (NETWARE_DEST_PATH "/share/postgres.bki", "w");
	fprintf (bkifile, "# PostgreSQL 7.2\n");
	fprintf (bkifile, "# 1 \"/tmp/genbkitmp.c\"\n");
	oidcounter = 0;
	printf ("Analyze pg_proc.h...\n");
	sprintf (bkiinputfile, (char *) "../../../include/catalog/pg_proc.h");
	nwgenbki (bkifile, bkiinputfile);
	printf ("Analyze pg_type.h...\n");
	sprintf (bkiinputfile, (char *) "../../../include/catalog/pg_type.h");
	nwgenbki (bkifile, bkiinputfile);
	printf ("Analyze pg_attribute.h...\n");
	sprintf (bkiinputfile, (char *) "../../../include/catalog/pg_attribute.h");
	nwgenbki (bkifile, bkiinputfile);
	printf ("Analyze pg_class.h...\n");
	sprintf (bkiinputfile, (char *) "../../../include/catalog/pg_class.h");
	nwgenbki (bkifile, bkiinputfile);
	printf ("Analyze pg_attrdef.h...\n");
	sprintf (bkiinputfile, (char *) "../../../include/catalog/pg_attrdef.h");
	nwgenbki (bkifile, bkiinputfile);
	printf ("Analyze pg_relcheck.h...\n");
	sprintf (bkiinputfile, (char *) "../../../include/catalog/pg_relcheck.h");
	nwgenbki (bkifile, bkiinputfile);
	printf ("Analyze pg_inherits.h...\n");
	sprintf (bkiinputfile, (char *) "../../../include/catalog/pg_inherits.h");
	nwgenbki (bkifile, bkiinputfile);
	printf ("Analyze pg_index.h...\n");
	sprintf (bkiinputfile, (char *) "../../../include/catalog/pg_index.h");
	nwgenbki (bkifile, bkiinputfile);
	printf ("Analyze pg_operator.h...\n");
	sprintf (bkiinputfile, (char *) "../../../include/catalog/pg_operator.h");
	nwgenbki (bkifile, bkiinputfile);
	printf ("Analyze pg_opclass.h...\n");
	sprintf (bkiinputfile, (char *) "../../../include/catalog/pg_opclass.h");
	nwgenbki (bkifile, bkiinputfile);
	printf ("Analyze pg_am.h...\n");
	sprintf (bkiinputfile, (char *) "../../../include/catalog/pg_am.h");
	nwgenbki (bkifile, bkiinputfile);
	printf ("Analyze pg_amop.h...\n");
	sprintf (bkiinputfile, (char *) "../../../include/catalog/pg_amop.h");
	nwgenbki (bkifile, bkiinputfile);
	printf ("Analyze pg_amproc.h...\n");
	sprintf (bkiinputfile, (char *) "../../../include/catalog/pg_amproc.h");
	nwgenbki (bkifile, bkiinputfile);
	printf ("Analyze pg_language.h...\n");
	sprintf (bkiinputfile, (char *) "../../../include/catalog/pg_language.h");
	nwgenbki (bkifile, bkiinputfile);
	printf ("Analyze pg_largeobject.h...\n");
	sprintf (bkiinputfile, (char *) "../../../include/catalog/pg_largeobject.h");
	nwgenbki (bkifile, bkiinputfile);
	oidcounter = 10000;
	printf ("Analyze pg_aggregate.h...\n");
	sprintf (bkiinputfile, (char *) "../../../include/catalog/pg_aggregate.h");
	nwgenbki (bkifile, bkiinputfile);
	oidcounter = 0;
	printf ("Analyze pg_statistic.h...\n");
	sprintf (bkiinputfile, (char *) "../../../include/catalog/pg_statistic.h");
	nwgenbki (bkifile, bkiinputfile);
	printf ("Analyze pg_rewrite.h...\n");
	sprintf (bkiinputfile, (char *) "../../../include/catalog/pg_rewrite.h");
	nwgenbki (bkifile, bkiinputfile);
	printf ("Analyze pg_trigger.h...\n");
	sprintf (bkiinputfile, (char *) "../../../include/catalog/pg_trigger.h");
	nwgenbki (bkifile, bkiinputfile);
	printf ("Analyze pg_listener.h...\n");
	sprintf (bkiinputfile, (char *) "../../../include/catalog/pg_listener.h");
	nwgenbki (bkifile, bkiinputfile);
	printf ("Analyze pg_description.h...\n");
	sprintf (bkiinputfile, (char *) "../../../include/catalog/pg_description.h");
	nwgenbki (bkifile, bkiinputfile);
	printf ("Analyze pg_database.h...\n");
	sprintf (bkiinputfile, (char *) "../../../include/catalog/pg_database.h");
	nwgenbki (bkifile, bkiinputfile);
	printf ("Analyze pg_shadow.h...\n");
	sprintf (bkiinputfile, (char *) "../../../include/catalog/pg_shadow.h");
	nwgenbki (bkifile, bkiinputfile);
	printf ("Analyze pg_group.h...\n");
	sprintf (bkiinputfile, (char *) "../../../include/catalog/pg_group.h");
	nwgenbki (bkifile, bkiinputfile);

	printf ("Analyze indexing.h...\n");
	sprintf (bkiinputfile, (char *) "../../../include/catalog/indexing.h");
	nwgenbki (bkifile, bkiinputfile);

	fprintf (bkifile ,"build indices\n");
	fclose (bkifile);
	return 0;
	}

void
RowSplitter (char* cvstring)
	{
	int		counter2		= 0;
	int		tokencounter	= 0;
	int		brblock			= 0;
	char	tokenstr[256];

	/* check for ascii 34 */
	newpghash = 0;

	brblock = 0;
	for (counter2 = 0; counter2 <= strlen (cvstring); counter2++)
		{
		if ((memchr	((char *) cvstring + counter2, 34, 1)) && (brblock == 0)) 
			brblock = 1;
			else if (memchr ((char *) cvstring + counter2, 34, 1))
			brblock = 0;

		if ((memchr	((char *) cvstring + counter2, 32, 1)) && (brblock == 1))
			memset	(((char *) cvstring + counter2), 95, 1);
		}

	for (counter2 = 0; counter2 <= strlen (cvstring); counter2++)
		{
		if (memchr	((char *) cvstring + counter2, 32, 1)
			|| memchr	((char *) cvstring + counter2, 9, 1))
			{
			memset	(((char *) tokenstr + tokencounter), 0, 1);
			WriteToken (tokenstr);
			tokencounter = 0;
			}
		else
			{
			memmove (tokenstr + tokencounter, cvstring + counter2, 1);
			tokencounter++;
			}
		}
	if (tokencounter)
		{
		memset	(((char *) tokenstr + tokencounter), 0, 1);
		WriteToken (tokenstr);
		}
	}

void
WriteToken (char* TokenString)
	{
	PGROWPTR	pghashtemp;
	PGROWPTR	pghashtempprev;

	// Analyze token and generate row
	if (strlen (TokenString) >= 11)
		if (strncmp ((char *) TokenString, "DATA(insert", 11) == 0) fmgrcounter = 1;

	if ((fmgrcounter > 0) && (strlen (TokenString) > 0))
		fmgrcounter++;

	if ((fmgrcounter == 5) && (strlen (TokenString) > 0))
		defnumber = atoi (TokenString);

	if ((fmgrcounter == 13) && (strlen (TokenString) > 0)) 
		{
		if (memchr	((char *) TokenString, 116, 1))
			defbool1 = 1;
			else
			defbool1 = 0;
		}

	if ((fmgrcounter == 14) && (strlen (TokenString) > 0)) 
		defvalue = atoi (TokenString);

	if ((fmgrcounter == 15) && (strlen (TokenString) > 0)) 
		{
		if (memchr	((char *) TokenString, 116, 1))
			defbool2 = 1;
			else
			defbool2 = 0;
		}

	if ((fmgrcounter == 22) && (strlen (TokenString) > 0)) 
		{
		if (memchr	((char *) TokenString, 34, 1) == 0)
			{
			pghashtemp		= pghashtable;
			pghashtempprev	= pghashtable;
			while (pghashtemp)
				{
				if (pghashtemp->number > defnumber)
					break;
				pghashtempprev = pghashtemp;
				pghashtemp = pghashtemp->next;
				}

			/* add new node */
			newpghash = malloc (sizeof(struct pgrow));
			newpghash->next			= pghashtempprev->next;
			newpghash->value		= defvalue;
			newpghash->number		= defnumber;
			newpghash->db1			= defbool1;
			newpghash->db2			= defbool2;
			memset (&newpghash->name[0], 0, 40);
			memcpy (&newpghash->name[0], (char *) TokenString, strlen (TokenString));
			pghashtempprev->next	= newpghash;
			fmgrcounter = 0;
			}
		}
	}

void
nwgenbki (FILE *bkihandle, char* inputfile)
	{
	int		stringloop		= 0;
	int		infile			= 0;
	int		statenumber		= 0;
	int		useful			= 0;
	int		counter			= 0;
	int		counter2		= 0;
	int		headline		= 0;
	int		crstatus1		= 0;
	int		crstatus2		= 0;
	int		crstatus3		= 0;
	int		tokenstate		= 0;
	int		tokencounter2	= 0;
	int		tokencounter3	= 0;
	int		badline			= 0;
	char	testlinestring[256];
	char	tokenstring[128];
	char	outputstring[256];
	char	testchar[1];
	char	part1[128];
	char	part2[128];
	char	part3[128];
	char	crpart1[128];
	char	crpart2[128];
	char	*s;

	infile = open (inputfile, O_RDONLY, 0);

	memset ((char *) part1, 0, 128);
	memset ((char *) part2, 0, 128);
	memset ((char *) part3, 0, 128);

	while (read (infile, testchar , 1) > 0)		/* read line of inputfile */
		{
		if (testchar[0] == 13)
			{
			memmove (testlinestring + stringloop, "\0", 1);
			read (infile, testchar , 1);

			if (strncmp ((char *) testlinestring, "CATALOG", 7) == 0)
				{
				useful		= 1;
				headline	= 1;
				for (counter2 = 0; counter2 <= strlen (testlinestring); counter2++)
					{
					if (memchr	((char *) testlinestring + counter2, 40, 1)) /* test for "(" */
						counter = counter2;

					if (memchr	((char *) testlinestring + counter2, 41, 1)) /* test for ")" */
						{
						memcpy ((char *) part1, (char *) testlinestring + counter + 1, (counter2 - counter - 1));
						memcpy ((char *) part2, (char *) testlinestring + counter2 + 1, strlen (testlinestring) - counter2 - 1);
						s = &part2[0];
						for (; *s; s++)	if (*s == 32)
							{
							s++;
							if (strncmp ((char *) s, "BKI_", 4) == 0)
								{
								s--;
								memset ((char *) s, 0, 1);
								s = s + 5;
								memcpy ((char *) part3, (char *) s, strlen (s));
								break;
								}
							}
						break;
						}
					}
				}

			if (strncmp ((char *) testlinestring, "DECLARE_UNIQUE_INDEX", 20) == 0)
				{
				fprintf (bkihandle, "declare unique index");
				s = &testlinestring[21];
				memset ((char *) s + strlen (s) - 2, 0 ,1);
				fprintf (bkihandle, " %s\n", s);
				}

			if (strncmp ((char *) testlinestring, "DECLARE_INDEX", 13) == 0)
				{
				fprintf (bkihandle, "declare index");
				s = &testlinestring[14];
				memset ((char *) s + strlen (s) - 2, 0 ,1);
				fprintf (bkihandle, " %s\n", s);
				}

			if (useful == 2)
				if (strncmp ((char *) testlinestring, "DATA(", 5) == 0)
					useful = 3;

			if (useful == 1)
				{
				if (strncmp ((char *) testlinestring, "{", 1) == 0)
					fprintf (bkihandle, " (\n");
				if (strncmp ((char *) testlinestring, "}", 1) == 0)
					{
					fseek(bkihandle, ftell(bkihandle) - 4, SEEK_SET);
					fprintf (bkihandle, "\n )\n");
					if (strlen (part2) < 8)
						fprintf (bkihandle, "open %s\n", part1);
					useful = 2;
					}

				if ((strlen (testlinestring) > 1) && (headline == 1))
					{
					headline = 0;
					s = &part2[0];
					for (; *s; s++)	if ((*s >= 65) && (*s <= 90)) *s += 32;
					s = &part3[0];
					for (; *s; s++)	if ((*s >= 65) && (*s <= 90)) *s += 32;
					if (strlen (part3) > 0)
						fprintf (bkihandle, "create%s %s %s\n", part2, part3, part1);
					else
						fprintf (bkihandle, "create%s %s\n", part2, part1);
					}
					else
					{
					if ((strlen (testlinestring) > 1) && (useful == 1))
						{
						crstatus1	= 0;
						crstatus2	= 0;
						crstatus3	= 0;
						memset (crpart1, 0, 128);
						memset (crpart2, 0, 128);
						badline		= 0;
						for (counter2 = 0; counter2 <= strlen (testlinestring); counter2++)
							{
							if (((memchr ((char *) testlinestring + counter2, 42, 1)) ||
							    (memchr ((char *) testlinestring + counter2, 47, 1))) &&
							    (crstatus1 == 0))
									{
									badline = 1;
									break;
									}

							if (!(memchr ((char *) testlinestring + counter2, 32, 1)) &&
							    !(memchr ((char *) testlinestring + counter2, 9, 1)) &&
							    (crstatus1 == 0))
									crstatus1 = counter2;

							if (((memchr ((char *) testlinestring + counter2, 32, 1)) ||
							    (memchr ((char *) testlinestring + counter2, 9, 1))) &&
							    (crstatus1 != 0) && (crstatus2 == 0))
									crstatus2 = counter2;
	
							if (!(memchr ((char *) testlinestring + counter2, 32, 1)) &&
							    !(memchr ((char *) testlinestring + counter2, 9, 1)) &&
							     (crstatus3 == 0) && (crstatus2 != 0))
									crstatus3 = counter2;

							if (((memchr ((char *) testlinestring + counter2, 32, 1)) ||
							    (memchr ((char *) testlinestring + counter2, 9, 1)) ||
							    (memchr ((char *) testlinestring + counter2, 59, 1))) &&
							    (crstatus3 != 0))
								{
								memcpy ((char *) crpart1, (char *) testlinestring + crstatus1, (crstatus2 - crstatus1));
								memcpy ((char *) crpart2, (char *) testlinestring + crstatus3, counter2 - crstatus3);
								s = &crpart1[0];
								for (; *s; s++)	if ((*s >= 65) && (*s <= 90)) *s += 32;
								s = &crpart2[0];
								for (; *s; s++)	if ((*s >= 65) && (*s <= 90)) *s += 32;

								if (strlen (crpart1) > 7)
									if (strncmp ((char *) crpart1, "namedata", 8) == 0)
										sprintf (crpart1, "name");

								if (strlen (crpart1) > 12)
									if (strncmp ((char *) crpart1, "transactionid", 13) == 0)
										sprintf (crpart1, "xid");

								if (strlen (crpart2) > 4)
									{
									if (strncmp ((char *) crpart2 + strlen (crpart2) - 1, "]", 1) == 0)
										{
										memset ((char *) crpart2 + strlen (crpart2) - 3, 0, 1);
										memset ((char *) crpart1 + strlen (crpart1), 91, 1);
										memset ((char *) crpart1 + strlen (crpart1), 93, 1);
										}
									}

								break;
								}
							}
						if (badline == 0) fprintf (bkihandle, " %s = %s ,\n", crpart2, crpart1);
						}
					}
				}

			if (useful == 3)
				{
				useful--;
				tokencounter2 = 0;
				tokencounter3 = 0;
				s = &testlinestring[5];
				for (; *s; s++)
					{
					if (*s > 32)
						{
						tokenstate = 1;
						memcpy ((char *) tokenstring + tokencounter2, &s[0] ,1);
						tokencounter2++;
						}
					if ((*s < 33) && (tokenstate == 1))
						{
						memset ((char *) tokenstring + tokencounter2, 32 ,1);
						tokencounter2++;
						memset ((char *) tokenstring + tokencounter2, 0 ,1);

						if ((tokencounter3 == 3) && (oidcounter > 0))
							{
							fprintf (bkihandle, "%d ", oidcounter);
							oidcounter++;
							}
							else
							{
							if (strncmp ((char *) tokenstring, "NAMEDATALEN", 11) == 0)
								fprintf (bkihandle, "%d ", namedatalen);

								else if (strncmp ((char *) tokenstring, "PGUID", 5) == 0)
									fprintf (bkihandle, "%d ", pguid);

								else if (strncmp ((char *) tokenstring, "ENCODING", 8) == 0)
									fprintf (bkihandle, "%d ", encoding);

								else if (strncmp ((char *) tokenstring, "INDEX_MAX_KEYS*2", 16) == 0)
									fprintf (bkihandle, "%d ", index_max_keys * 2);

								else if (strncmp ((char *) tokenstring, "INDEX_MAX_KEYS*4", 16) == 0)
									fprintf (bkihandle, "%d ", index_max_keys * 4);

								else if (strncmp ((char *) tokenstring, "INDEX_MAX_KEYS", 14) == 0)
									fprintf (bkihandle, "%d ", index_max_keys);

								else if (strncmp ((char *) tokenstring, "DEFAULT_ATTSTATTARGET", 21) == 0)
									fprintf (bkihandle, "%d ", def_attstattarget);

								else fprintf (bkihandle, "%s", tokenstring);
							}
						tokencounter3++;
						tokencounter2 = 0;
						tokenstate = 0;
						}
					}

				if ((oidcounter == 0) || (tokencounter3 != 3))
					{
					memset ((char *) tokenstring + tokencounter2 - 2, 0 ,1);
					fprintf (bkihandle, "%s\n", tokenstring);
					}
				}
			stringloop = 0;
			}
			else
			{
			memmove (testlinestring + stringloop, testchar, 1);
			stringloop++;
			}
		}

	if (strlen (part1) > 1) fprintf (bkihandle, "close %s\n", part1);
	close (infile);
	}