/*------------------------------------------------------------------------------
 *
 * $name: os.c
 * $version: 0.65
 * $date_modified: 2002/07/22 02:00:00
 * $description: this file contains definitions for postgres on NetWare
 * $written_by: Ulrich Neumann
 * $last_modified_by: Ulrich Neumann
 *------------------------------------------------------------------------------
 */
#include "os.h"

#include <sys/sem.h>
#include <stdio.h>
#include <termios.h>

struct semid_ds NWSemid_ds[101];

extern QueryContext;
extern CacheMemoryContext;
extern PostmasterContext;
extern ErrorContext;
extern TopMemoryContext;
extern fscxt;
extern DataDir;
extern BackendList;

char	**nwnew_argv;
int		nwargc;

/* BEGIN POSIX USER/GROUP FUNCTIONS */
passwd*
getpwuid (int userid)
	{
	/* later we can add some LDAP stuff here */
	pws->pw_name = (char *) "postgres\0";
	pws->pw_passwd = (char *) "\0";
	pws->pw_uid = 1;
	pws->pw_gid = 1;
	pws->pw_gecos = (char *) "\0";
	pws->pw_dir = (char *) "/\0";
	pws->pw_shell = (char *) "bash\0";
	return pws;
	}

group*
getgrgid (gid_t groupid)
	{
	grs->gr_name = (char *) "everyone\0";
	grs->gr_passwd = (char *) "\0";
	grs->gr_gid = 1;
	return grs;
	}


int getppid()
{
	return main_thread_id;
}

/* END POSIX USER/GROUP FUNCTIONS, maybe obsolete with future LibC */


/* begin NetWare specific functions */
void
NWInit(int argc, char *argv[])			/* init NetWare specific values			 */
	{
	int		loopctr = 0;
	int		flag	= 0;

	int		filedes	= 0;
	struct	termios mytp;
	int		retval	= 0;

	/* don't auto close screen */

	nwnew_argv = argv;
	nwargc = argc;

	if (argc > 1)
	{
		if (strcmp(argv[1], "--help") == 0 || strcmp(argv[1], "-?") == 0 ||
		    strcmp(argv[1], "?") == 0 || strcmp(argv[1], "--version") == 0 ||
		    strcmp(argv[1], "-V") == 0)
			setscreenmode (0);
	}

	if (argc > 2)
	{
		if (strcmp(argv[2], "--help") == 0 || strcmp(argv[2], "-?") == 0 ||
		    strcmp(argv[2], "?") == 0 || strcmp(argv[2], "--version") == 0 ||
		    strcmp(argv[2], "-V") == 0)
			setscreenmode (0);
	}

//	setscreenmode (0);

	main_thread_id	= pthread_self();

	atexit (NWCleanUp);

	/* Block CTRL-C in interactive backend */
	retval = tcgetattr(filedes = fileno(stdin), &mytp);
	mytp.c_cc[VINTR] = 1;
	mytp.c_cc[VEOF] = 0;
	mytp.c_cc[VSUSP] = 1;
	retval = tcsetattr(filedes, TCSANOW, &mytp);

	/* initialize semaphores */
	for (loopctr = 0; loopctr <= 100; loopctr++)
		NWSemid_ds[loopctr].sem_base = 0;

	/* initialize shared memory */
	for (loopctr = 0; loopctr <= 100; loopctr++)
		NWPGMem[loopctr] = 0;

	/* initialize struct group */
	grs = (struct group *) malloc (sizeof (struct group));

	/* initialize struct passwd */
	pws = (struct passwd *) malloc (sizeof (struct passwd));
	}

void
NWCleanUp(argc, argv)			/* free NetWare specific values			*/
{
	int loopctr = 0;

	/* Release grs and pws struct's */
	free(grs);
	free(pws);

	/* free strdup of environment */
	for (loopctr = 0; loopctr < nwargc; loopctr++)
		if (nwnew_argv[loopctr]) free(nwnew_argv[loopctr]);
	free(nwnew_argv);

	/* Clean up main memory contexts */
	if (fscxt)
		MemoryContextDelete(fscxt);

	if (QueryContext)
		MemoryContextDelete(QueryContext);

	if (CacheMemoryContext)
		MemoryContextDelete(CacheMemoryContext);
	
	if (PostmasterContext)
		MemoryContextDelete(PostmasterContext);

	if (ErrorContext)
		MemoryContextDelete(ErrorContext);

	if (TopMemoryContext)
		MemoryContextDelete(TopMemoryContext);

	DLFreeList(BackendList);

	free (DataDir);
}
/* end NetWare specific functions */


/* Set the timer WHICH to *NEW.  If OLD is not NULL,
   set *OLD to the old value of timer WHICH.
   Returns 0 on success, -1 on errors.  */
long
setitimer(int which, struct itimerval *newvalue,
		  struct itimerval *oldvalue)
{
	if (newvalue)
	{
		NXThreadDelay(1000);
	}
	return 0;
}


void
sync ()							/* dummy function						*/
{
	return;
}


int
waitpid (int pid, int *stat_loc, int options)
{
	int wpcounter		= 0;
	int	retpid			= 0;

	for (wpcounter = 0; wpcounter < 32; wpcounter++)
	{
		if (NWwaitpidtable[wpcounter].thread_id != 0)
		{
			retpid = NWwaitpidtable[wpcounter].thread_id;
			NWwaitpidtable[wpcounter].thread_id = 0;

			*stat_loc = NWwaitpidtable[wpcounter].exitstatus;

			return retpid;
		}
	}
	return retpid;
}

/* ----------------------------------------------------
 * emulate SYS-V IPC semaphores on NetWare
 * ----------------------------------------------------
 */

int
semget (int semKey, int semNum, int flag)
{
	int loopctr = 0;
	int semctr	= 0;

	for (loopctr = 1; loopctr <= 100; loopctr++)
		if (NWSemid_ds[loopctr].sem_base == 0) break;

	if (loopctr > 100) return -1;

	if (semKey == 0) semKey = loopctr;

	if (NWSemid_ds[loopctr].sem_base == 0)
	{
		NWSemid_ds[loopctr].sem_base = (struct sem*) NXMemAlloc(sizeof(struct sem) * semNum, 0);
		NWSemid_ds[loopctr].sem_nsems = semNum;
		NWSemid_ds[loopctr].sem_otime = 0;
		NWSemid_ds[loopctr].base_nxsem_ptr = NXSemaAlloc(1, NULL);
		NWSemid_ds[loopctr].sem_internal_key = semKey;

		for (semctr = 0; semctr < semNum; semctr++)
		{
			NWSemid_ds[loopctr].sem_base[semctr].nxsem_ptr = NXSemaAlloc(1, NULL);
		}
		NWSemid_ds[loopctr].sem_ctime = time(NULL);
	}
	return semKey;
}


int
semctl (int semId, int semNum, int flags, /* ... */ union semun arg)
{
	int retval			= 0;
	int semCount		= 0;
	int	loopctr			= 0;

	for (loopctr = 1; loopctr <= 100; loopctr++)
		if (NWSemid_ds[loopctr].sem_internal_key == semId) break;

	semId = loopctr;

	if (NWSemid_ds[semId].base_nxsem_ptr == NULL)
	{
		errno = EINVAL;
		return -1;
	}

	NXSemaWait (NWSemid_ds[semId].base_nxsem_ptr);

	if (semId < 0 || semId >= 101 ||
		semNum < 0 || semNum >= NWSemid_ds[semId].sem_nsems)
	{
		NXSemaPost (NWSemid_ds[semId].base_nxsem_ptr);
		errno = EINVAL;
		return -1;
	}

	switch (flags)
	{
		case GETNCNT:
			retval = NWSemid_ds[semId].sem_base[semNum].semncnt;
			break;

		case GETPID:
			retval = NWSemid_ds[semId].sem_base[semNum].sempid;
			break;

		case GETVAL:
			retval = NWSemid_ds[semId].sem_base[semNum].semval;
			break;

		case GETALL:
			for (semCount = 0; semCount < NWSemid_ds[semId].sem_nsems; semCount++)
				arg.array[semCount] = NWSemid_ds[semId].sem_base[semCount].semval;
			break;

		case SETVAL:
			NWSemid_ds[semId].sem_base[semNum].semval = arg.val;
			break;

		case SETALL:
			for (semCount = 0; semCount < NWSemid_ds[semId].sem_nsems; semCount++)
				NWSemid_ds[semId].sem_base[semCount].semval = arg.array[semCount];
			break;

		case GETZCNT:
			retval = NWSemid_ds[semId].sem_base[semNum].semzcnt;
			break;

		case IPC_RMID:
			NXSemaPost (NWSemid_ds[semId].base_nxsem_ptr);
			NXSemaFree (NWSemid_ds[semId].base_nxsem_ptr);

			for (semCount = 0; semCount < NWSemid_ds[semId].sem_nsems; semCount++)
				NXSemaFree (NWSemid_ds[semId].sem_base[semCount].nxsem_ptr);

			NXMemFree (NWSemid_ds[semId].sem_base);
			NWSemid_ds[semId].sem_base = 0;
			NWSemid_ds[semId].sem_nsems = 0;
			return 0;
			break;

		case IPC_STAT:
			memcpy (arg.buf, &NWSemid_ds[semId], sizeof(struct semid_ds));
			retval = NWSemid_ds[loopctr].sem_internal_key;
			break;

		default:
			NXSemaPost (NWSemid_ds[semId].base_nxsem_ptr);
			errno = EINVAL;
			return -1;
	}

	NXSemaPost (NWSemid_ds[semId].base_nxsem_ptr);

	return retval;
}


int
semop(int semId, struct sembuf * sops, size_t nsops)
{
	int		i;
	int		op;
	int		retval;
	int	loopctr			= 0;

	for (loopctr = 1; loopctr <= 100; loopctr++)
		if (NWSemid_ds[loopctr].sem_internal_key == semId) break;

	semId = loopctr;

	if (NWSemid_ds[semId].base_nxsem_ptr == NULL)
	{
		errno = EINVAL;
		return -1;
	}

	NXSemaWait (NWSemid_ds[semId].base_nxsem_ptr);

	if ((semId < 0) || (semId >= NWSemid_ds[semId].sem_nsems))
	{
		NXSemaPost (NWSemid_ds[semId].base_nxsem_ptr);
		errno = EINVAL;
		return -1;
	}

	for (i = 0; i < nsops; i++)
	{
		/* For each sem in the pool, check the operation to perform */
		if (sops[i].sem_op < 0) /* LOCK and TRYLOCK */
		{
			if (NWSemid_ds[semId].sem_base[sops[i].sem_num].semval < -sops[i].sem_op)
			{
				if (sops[i].sem_flg & IPC_NOWAIT)
				{
					NXSemaPost (NWSemid_ds[semId].base_nxsem_ptr);
				printf("We are in IPC_NOWAIT failure");
					errno = EAGAIN;
					return -1;
				}

				NWSemid_ds[semId].sem_base[sops[i].sem_num].semncnt++;

				NXSemaPost (NWSemid_ds[semId].base_nxsem_ptr);
				NXSemaWait (NWSemid_ds[semId].sem_base[sops[i].sem_num].nxsem_ptr);
				NXSemaWait (NWSemid_ds[semId].base_nxsem_ptr);

				NWSemid_ds[semId].sem_base[sops[i].sem_num].semval -= -sops[i].sem_op;

				if (sops[i].sem_flg & SEM_UNDO)
				{
					NWSemid_ds[semId].sem_base[sops[i].sem_num].semadj += -sops[i].sem_op;
				}
				NWSemid_ds[semId].sem_base[sops[i].sem_num].semncnt--;
			}
			else
			{
				NWSemid_ds[semId].sem_base[sops[i].sem_num].semval -= -sops[i].sem_op;
				if (sops[i].sem_flg & SEM_UNDO)
				{
					NWSemid_ds[semId].sem_base[sops[i].sem_num].semadj += -sops[i].sem_op;
				}
			}

		}
		else if (sops[i].sem_op > 0) /* UNLOCK/POST */
		{
			NWSemid_ds[semId].sem_base[sops[i].sem_num].semval += sops[i].sem_op;

			if (sops[i].sem_flg & SEM_UNDO)
			{
				NWSemid_ds[semId].sem_base[sops[i].sem_num].semadj -= sops[i].sem_op;
			}
			NXSemaPost(NWSemid_ds[semId].sem_base[sops[i].sem_num].nxsem_ptr);
		}
		else /* sem_op == 0 not used */
		{
		}
		NWSemid_ds[semId].sem_base[sops[i].sem_num].sempid = NXVmGetId();

	}
	NWSemid_ds[semId].sem_otime = time(NULL);
	NXSemaPost (NWSemid_ds[semId].base_nxsem_ptr);

	return 0;
}


/* -------------------------------------------------------
 * emulate SYS-V style shared memory on NetWare for pgsql
 * (very thin emulation !!)
 * -------------------------------------------------------
 */

int
shmdt(char *shmaddr)
{
	int retval = -1;
	int loopctr;

	for (loopctr = 1; loopctr <= 100; loopctr++)
	{
		if (((char *) shmaddr - NWPGMem[loopctr]) >= 0 &&
			((char *) shmaddr - NWPGMem[loopctr]) <= (NWShmid_ds[loopctr].shm_segsz))
		{
			if (NWShmid_ds[loopctr].shm_nattch > 0)
			{
				NWShmid_ds[loopctr].shm_nattch--;
				NXGetTime(NX_SINCE_BOOT, NX_USECONDS, &NWShmid_ds[loopctr].shm_dtime);
				retval = 0;
				break;
			}
		}
	}
	return retval;
}

void* 
shmat(int shmid, void *shmaddr, int flag)
{
	int loopctr;
	for (loopctr = 1; loopctr <= 100; loopctr++)
	{
		if (NWPGMem[loopctr] != 0)
		{
			if (NWShmid_ds[loopctr].shm_internal_key == shmid)
			{
				shmid = loopctr;
				break;
			}
		}
	}

	if (shmid <= 100)
	{
		if (NWPGMem[shmid] > (void *) 0)
		{
			NWShmid_ds[shmid].shm_nattch++;
			NXGetTime(NX_SINCE_BOOT, NX_USECONDS, &NWShmid_ds[shmid].shm_atime);
			return NWPGMem[shmid];
		}
	}
	return (int *) -1;
}

int
shmctl(int shmid, int flag, struct shmid_ds * shmemtemp)
{
	int loopctr;
	int retval = -1;

	for (loopctr = 1; loopctr <= 100; loopctr++)
	{
		if (NWPGMem[loopctr] != 0)
		{
			if (NWShmid_ds[loopctr].shm_internal_key == shmid)
			{
				shmid = loopctr;
				break;
			}
		}
	}

	switch (flag)
	{
		case (IPC_RMID):
			NXMemFree (NWPGMem[shmid]);
			NWPGMem[shmid] = 0;
			retval = 0;
			break;

		case (IPC_STAT):
			memcpy (shmemtemp, &NWShmid_ds[shmid].shm_segsz, sizeof (struct shmid_ds));
			retval = 0;
			break;

		case (IPC_SET):
			NXGetTime(NX_SINCE_BOOT, NX_USECONDS, &NWShmid_ds[shmid].shm_ctime);
			memcpy (shmemtemp, &NWShmid_ds[shmid].shm_segsz, sizeof (struct shmid_ds));
			break;

		case (SHM_LOCK):
			NXMemCtl(NWPGMem[shmid], NWShmid_ds[shmid].shm_segsz, NX_MEM_LOCK);
			NXGetTime(NX_SINCE_BOOT, NX_USECONDS, &NWShmid_ds[shmid].shm_ctime);
			memcpy (shmemtemp, &NWShmid_ds[shmid].shm_segsz, sizeof (struct shmid_ds));
			break;

		case (SHM_UNLOCK):
			NXMemCtl(NWPGMem[shmid], NWShmid_ds[shmid].shm_segsz, NX_MEM_UNLOCK);
			NXGetTime(NX_SINCE_BOOT, NX_USECONDS, &NWShmid_ds[shmid].shm_ctime);
			memcpy (shmemtemp, &NWShmid_ds[shmid].shm_segsz, sizeof (struct shmid_ds));
			break;

		default:
			break;
	}
 
	return retval;
}

int
shmget(int key, int size, int flag)
{
	int loopctr = 0;

	switch (flag)
	{
		case IPC_PRIVATE:
			for (loopctr = 1; loopctr <= 100; loopctr++)
				if (NWPGMem[loopctr] == 0) break;

			if (loopctr > 100) return -1;
			if (key <= 0) key = loopctr;

			NWPGMem[loopctr] = NXMemAlloc (size, 0);
			memset (NWPGMem[loopctr], 0, size);
			NWShmid_ds[loopctr].shm_segsz			= size;
			NWShmid_ds[loopctr].shm_lkcnt			= 0;
			NWShmid_ds[loopctr].shm_lpid			= 0;
			NWShmid_ds[loopctr].shm_cpid			= NXThreadGetId();
			NWShmid_ds[loopctr].shm_nattch			= 0;
			NWShmid_ds[loopctr].shm_cnattch			= 0;
			NWShmid_ds[loopctr].shm_atime			= 0;
			NWShmid_ds[loopctr].shm_pad1			= 0;
			NWShmid_ds[loopctr].shm_dtime			= 0;
			NWShmid_ds[loopctr].shm_pad2			= 0;
			NXGetTime(NX_SINCE_BOOT, NX_USECONDS, &NWShmid_ds[loopctr].shm_ctime);
			NWShmid_ds[loopctr].shm_pad3			= 0;
			NWShmid_ds[loopctr].shm_internal_key	= key;

			return key;
			break;

		default:
			// Create new memory segment
			for (loopctr = 1; loopctr <= 100; loopctr++)
				if (NWPGMem[loopctr] != 0)
					if ((NWShmid_ds[loopctr].shm_internal_key == key) && (key != 0)) return key;

			for (loopctr = 1; loopctr <= 100; loopctr++)
				if (NWPGMem[loopctr] == 0) break;

			if (loopctr > 100) return -1;
			if (key <= 0) key = loopctr;

			NWPGMem[loopctr] = NXMemAlloc (size, 0);
			memset (NWPGMem[loopctr], 0, size);
			NWShmid_ds[loopctr].shm_segsz			= size;
			NWShmid_ds[loopctr].shm_lkcnt			= 0;
			NWShmid_ds[loopctr].shm_lpid			= 0;
			NWShmid_ds[loopctr].shm_cpid			= NXThreadGetId();
			NWShmid_ds[loopctr].shm_nattch			= 0;
			NWShmid_ds[loopctr].shm_cnattch			= 0;
			NWShmid_ds[loopctr].shm_atime			= 0;
			NWShmid_ds[loopctr].shm_pad1			= 0;
			NWShmid_ds[loopctr].shm_dtime			= 0;
			NWShmid_ds[loopctr].shm_pad2			= 0;
			NXGetTime(NX_SINCE_BOOT, NX_USECONDS, &NWShmid_ds[loopctr].shm_ctime);
			NWShmid_ds[loopctr].shm_pad3			= 0;
			NWShmid_ds[loopctr].shm_internal_key	= key;

			return loopctr;
	}
	return -1;
}
