/*------------------------------------------------------------------------------
 *
 * $name: os.h
 * $version: 1.01
 * $date_modified: 2002/01/12 00:17:00
 * $description: this file contains definitions for postgres on NetWare
 * $written_by: Ulrich Neumann
 * $last_modified_by: Ulrich Neumann
 *
 *------------------------------------------------------------------------------
 */

#ifndef N_PLAT_NLM_OS_DEF
#define N_PLAT_NLM_OS_DEF

#define bool char	/* needs to be defined before any header will be included */

#define N_PLAT_NLM 1
#define DLSUFFIX ".NLM"
#define PKGLIBDIR "SYS:/postgres/modules"

#include <unistd.h>
#include <errno.h>
#include <sys/select.h>
#include <sys/time.h>
#include <dlfcn.h>
#include <pwd.h>
#include <grp.h>
#include <nksapi.h>
#include <pthread.h>


#define SIGWINCH		  40
#define SA_RESTART   0x10000000 		/* Restart syscall on signal return. 		*/

#define IPC_CREAT	 0001000			/* create key if key does not exist.		*/
#define IPC_EXCL	 0002000			/* fail if key exists.						*/
#define IPC_NOWAIT	 0004000			/* return error on wait.					*/
#define IPC_PRIVATE	(key_t)0			/* create private memory segment			*/

/* control commands for 'semctl', and 'shmctl'.	*/
#define IPC_RMID		   0			/* remove identifier.						*/
#define IPC_STAT		   1			/* get `ipc_perm' options.					*/
#define IPC_SET			   2			/* set `ipc_perm' options.					*/
#define IPC_INFO		   3			/* see ipcs.								*/
#define SHM_LOCK		   4			/* lock memory								*/
#define SHM_UNLOCK		   5			/* unlock memory							*/

#define	WNOHANG			   1			/* don't block wait for pid.				*/
#define	WUNTRACED		   2			/* report status of stopped children		*/

/* Nonzero if STATUS indicates normal termination.  */
#define	WIFEXITED(status)	((status) == 0)
#define WEXITSTATUS(status)	(status)

#define TCSADRAIN		   2			/* not yet implemented in termios.h         */
#define TCSAFLUSH		   3			/* not yet implemented in termios.h         */

#define kill		NXThreadInterrupt

/* begin global variables */
struct	group *grs;
struct	passwd *pws;
int		main_thread_id;
int		myInternalEnv;

/* begin stuff that is planned to be obsolete with future versions of LibC */
struct itimerval
{
	struct timeval it_interval;
	struct timeval it_value;			/* Time to the next timer expiration.		*/
};

struct itimerspec
{
	struct timespec it_interval;
	struct timespec it_value;
};

#define crypt(password, salt) (password)
typedef unsigned short ushort;
#define ITIMER_REAL		   1
void	sync(void);								/* dummy function					*/
/* end stuff that is planned to be obsolete with future versions of LibC */


/* begin function prototypes */
void	NWInit		(int argc, char *argv[]);	/* init Netware specific stuff		*/
void	NWCleanUp	(void);						/* clean up							*/
int		shmdt		(char *shmaddr);
void	*shmat		(int shmid, void *shmaddr, int flag);
int		shmctl		(int shmid, int flag, struct shmid_ds * shmemtemp);
int		shmget		(int key, int size, int flag);
long	setitimer	(int which, struct itimerval *newvalue,
					 struct itimerval *oldvalue);
int		getppid		(void);
/* end function prototypes */


/* 
 * System V IPC shared memory emulation for NetWare.
 */
void    *NWPGMem[101];

struct shmid_ds
{
	int				shm_segsz;
	int				dummy;
	unsigned short	shm_lkcnt;			/* number of times it is locked				*/
	int				shm_lpid;			/* pid of last shmop						*/
	int				shm_cpid;			/* pid of creator							*/
	int				shm_nattch;			/* only for shminfo							*/
	int				shm_cnattch;		/* only for shminfo							*/
	uint64_t		shm_atime;			/* last shmat time							*/
	uint64_t		shm_pad1;			/* reserved for time_t expansion			*/
	uint64_t		shm_dtime;			/* last shmdt time							*/
	uint64_t		shm_pad2;			/* reserved for time_t expansion			*/
	uint64_t		shm_ctime;			/* last change time							*/
	uint64_t		shm_pad3;			/* reserved for time_t expansion			*/
	uint64_t		shm_pad4[4];		/* reserved area							*/
	int				shm_internal_key;	/* 'real' shared memory key					*/
};

struct shmid_ds NWShmid_ds[100];

/*
 * Table for thread based wait() emulation
 */
struct waitpidtable
{
	int				thread_id;
	int				exitstatus;
};

struct waitpidtable NWwaitpidtable[32];

#endif
